var SmPrefs = {
  prefs: Components.classes["@mozilla.org/preferences-service;1"]
         .getService(Components.interfaces.nsIPrefService)
         .getBranch("extensions.simplemail."),

  getIntPref: function(name) {
    try {
      return this.prefs.getIntPref(name);
    } catch(e) {
      return 0;
    }
  },

  setIntPref: function(name, value) {
    this.prefs.setIntPref(name, value);
  },

  getCharPref: function(name) {
    try {
      return this.prefs.getComplexValue(name,
        Components.interfaces.nsISupportsString).data;
    } catch(e) {
      return null;
    }
  },

  setCharPref: function(name, value) {
    var str = Components.classes["@mozilla.org/supports-string;1"]
              .createInstance(Components.interfaces.nsISupportsString);
    str.data = value;
    this.prefs.setComplexValue(name,
      Components.interfaces.nsISupportsString, str);
  },

  getBoolPref: function(name) {
    try {
      return this.prefs.getBoolPref(name);
    } catch(e) {
      // Old versions of SM stored Bool prefs as Ints. Convert old value, if exists.
      try {
        var value = this.prefs.getIntPref(name) == 1;
        this.deletePref(name);
        this.prefs.setBoolPref(name, value);
        return value;
      } catch(e) {
        return false;
      }
    }
  },

  setBoolPref: function(name, value) {
    this.prefs.setBoolPref(name, value);
  },

  getObjectPref: function(name) {
    return eval(this.getCharPref(name));
  },

  setObjectPref: function(name, value) {
    this.setCharPref(name, SmUtils.toJSON(value));
  },

  deletePref: function(name) {
    try {
      this.prefs.deleteBranch(name);
    } catch(e) {}
  },

  clearPref: function(name) {
    try {
      this.prefs.clearUserPref(name);
    } catch(e) {}
  }
}

var SmFile = {
  getDataDirectory: function() {
    var file;
    try {
      file = SmPrefs.prefs.getComplexValue("dataDirectory",
                                           Components.interfaces.nsIRelativeFilePref).file;
    }
    catch(e) {
      file = SmPrefs.prefs.getComplexValue("dataDirectory",
                                           Components.interfaces.nsILocalFile);
    }
    if (!file.exists()) file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
    return file;
  },

  // Unlimited number of arguments
  getDirectory: function() {
    var dir = SmFile.getDataDirectory();
    for(var i = 0; i < arguments.length; i++) {
      dir.append(arguments[i]);
      if (!dir.exists()) dir.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
    }
    return dir;
  },

  // Unlimited nuber of arguments
  getURL: function() {
    var url = "";
    for(var i = 0; i < arguments.length; i++) {
      url += url ? "/" + arguments[i] : arguments[i];
    }
    return encodeURI("simplemail://" + url);
  },

  getAttachmentsDirectory: function(directory) {
    return SmFile.getDirectory("attachments", directory);
  },

  saveAttachment: function(directory, name, contents) {
    var file = SmFile.getAttachmentsDirectory(directory);
    file.append(name);
    SmFile.writeFile(file, contents);
  },

  deleteAttachments: function(directory) {
    var file = SmFile.getAttachmentsDirectory(directory);
    if (file.exists()) {
      try { file.remove(true); } catch(e) { SmUtils.error("cantRemoveAttachmentsDir"); }
    }
  },

  getAttachmentURL: function(directory, name) {
    return SmFile.getURL("attachments", directory, name);
  },

  getImagesDirectory: function() {
    return SmFile.getDirectory("images");
  },

  deleteImage: function(name) {
    var file = SmFile.getImagesDirectory();
    file.append(name);
    try { file.remove(false); } catch(e) { SmUtils.error("cantRemoveImage", [name]); }
  },

  getImageURL: function(name) {
    return SmFile.getURL("images", name);
  },

  getSoundsDirectory: function() {
    return SmFile.getDirectory("sounds");
  },

  deleteSound: function(name) {
    var file = SmFile.getSoundsDirectory();
    file.append(name);
    try { file.remove(false); } catch(e) { SmUtils.error("cantRemoveSound", [name]); }
  },

  getSoundURL: function(name) {
    return SmFile.getURL("sounds", name);
  },

  saveURL: function(url, file) {
    var io = Components.classes["@mozilla.org/network/io-service;1"]
                       .getService(Components.interfaces.nsIIOService);
    var source = io.newURI(url, "UTF-8", null);
      
    var persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                            .createInstance(Components.interfaces.nsIWebBrowserPersist);
    
    persist.persistFlags = Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
    persist.persistFlags |= Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;
    
    persist.saveURI(source, null, null, null, null, file);
  },

  saveCanvas: function(canvas, fileName) {
    var file = SmFile.getImagesDirectory();
    file.append(fileName);
    SmFile.saveURL(canvas.toDataURL("image/png"), file);
  },

  createOutputStream: function(file) {
    var stream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                 .createInstance(Components.interfaces.nsIFileOutputStream);
    stream.init(file, 0x04 | 0x08 | 0x20, 0666, 0); // read & write, create, truncate
    var buffered = Components.classes["@mozilla.org/network/buffered-output-stream;1"]
                   .createInstance(Components.interfaces.nsIBufferedOutputStream);
    buffered.init(stream, 1*1024*1024);
    return { stream: stream,
             buffered: buffered,
             write: function(data) {
               this.buffered.write(data, data.length);
             },
             close: function() {
               this.buffered.close();
               this.stream.close();
             }
           };
  },

  createInputStream: function(file) {
    var stream = Components.classes["@mozilla.org/network/file-input-stream;1"]
                 .createInstance(Components.interfaces.nsIFileInputStream);
    stream.init(file, -1, 0, 0);
    var binary = Components.classes["@mozilla.org/binaryinputstream;1"]
                 .createInstance(Components.interfaces.nsIBinaryInputStream);
    binary.setInputStream(stream);
    return { stream: stream,
             binary: binary,
             readFully: function() {
               return this.binary.readBytes(this.binary.available());
             },
             close: function() {
               this.binary.close();
               this.stream.close();
             }
           };
  },

  getNsiFile: function(file) {
    if (file instanceof Components.interfaces.nsIFile) return file;
    else {
      var nsiFile = Components.classes["@mozilla.org/file/local;1"]
                .createInstance(Components.interfaces.nsILocalFile);
      nsiFile.initWithPath(file);
      return nsiFile;
    }
  },

  readFile: function(file, unichar) {
    var nsiFile = SmFile.getNsiFile(file);
    var inn = unichar ? SmFile.createUnicharInputStream(nsiFile) 
                      : SmFile.createInputStream(nsiFile);
    try {
      return inn.readFully();
    }
    finally {
      inn.close();
    }
  },

  writeFile: function(file, data, unichar) {
    var nsiFile = SmFile.getNsiFile(file);
    var out = unichar ? SmFile.createUnicharOutputStream(nsiFile)
                      : SmFile.createOutputStream(nsiFile);
    try {
      out.write(data);
    }
    finally {
      out.close();
    }
  },

  createTemporaryFile: function() {
    var file = Components.classes["@mozilla.org/file/directory_service;1"]
               .getService(Components.interfaces.nsIProperties)
               .get("TmpD", Components.interfaces.nsIFile);
    file.append("simplemail.tmp");
    file.createUnique(file.NORMAL_FILE_TYPE, 0666);

    var out = SmFile.createOutputStream(file);
    var inn = SmFile.createInputStream(file);

    return { write: function(data) { out.write(data); },
             readFully: function() {
               out.buffered.flush();
               return inn.readFully();
             },
             remove: function() {
               out.close();
               inn.close();
               file.remove(false);
             }
           };
  },

  createUnicharOutputStream: function(file) {
    var stream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                 .createInstance(Components.interfaces.nsIFileOutputStream);
    stream.init(file, 0x04 | 0x08 | 0x20, 0666, 0); // read & write, create, truncate
    var unichar = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                  .createInstance(Components.interfaces.nsIConverterOutputStream);
    unichar.init(stream, "utf-8", 0, 0xFFFD);

    return { stream: stream,
             unichar: unichar,
             write: function(data) {
               this.unichar.writeString(data);
             },
             close: function() {
               this.unichar.close();
               this.stream.close();
             }
           };
  },

  createUnicharInputStream: function(file) {
    var stream = Components.classes["@mozilla.org/network/file-input-stream;1"]
                 .createInstance(Components.interfaces.nsIFileInputStream);
    stream.init(file, -1, 0, 0);
    var unichar = Components.classes["@mozilla.org/intl/converter-input-stream;1"]
                  .createInstance(Components.interfaces.nsIConverterInputStream);
    unichar.init(stream, "utf-8", stream.available(), 0xFFFD);
    return { stream: stream,
             unichar: unichar,
             readFully: function() {
               var str = {};
               this.unichar.readString(this.stream.available(), str);
               return str.value;
             },
             close: function() {
               this.unichar.close();
               this.stream.close();
             }
           };
  },

  getStatusImageURL: function(status) {
    var image = (status == "loading") ? "loading.gif" : status + ".png";
    return "chrome://simplemail/skin/status/" + image;
  },

  getFileURL: function(file) {
    var ios = Components.classes["@mozilla.org/network/io-service;1"]
              .getService(Components.interfaces.nsIIOService);
    var fileHandler = ios.getProtocolHandler("file")
                     .QueryInterface(Components.interfaces.nsIFileProtocolHandler);
    return fileHandler.getURLSpecFromFile(file);
  },

  getMimeType: function(uriOrFile) {
    try {
      var mimeService = Components.classes["@mozilla.org/mime;1"]
                        .getService(Components.interfaces.nsIMIMEService);
      if (uriOrFile instanceof Components.interfaces.nsIFile) {
        return mimeService.getTypeFromFile(uriOrFile);
      } else {
        return mimeService.getTypeFromURI(uriOrFile);
      }
    } catch(e) {
      return "application/octet-stream";
    }
  },

  readURL: function(url) {
    var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
    var bstream = Components.classes["@mozilla.org/binaryinputstream;1"]
                  .createInstance(Components.interfaces.nsIBinaryInputStream);
    var channel = ioService.newChannel(url, null, null);
    var input = channel.open();
    bstream.setInputStream(input);
    var contents = bstream.readBytes(bstream.available());
    bstream.close();
    input.close();
    return contents;
  },

  getFileName: function(url) {
    var nsiUrl = (nsiUrl instanceof Components.interfaces.nsIURL)
                   ? url : SmFile.getNsiURL(url);
    return decodeURIComponent(nsiUrl.fileName);
  },

  getNsiURL: function(url) {
    var nsiUrl = Components.classes["@mozilla.org/network/standard-url;1"]
               .createInstance(Components.interfaces.nsIURL);
    nsiUrl.spec = url;
    return nsiUrl;
  }
}
