const BOOL = 1, INT = 2, CHAR = 3;

var SmExport = {

  prefs: { checkOnStartup: BOOL,
           deleteFromServer: BOOL,
           errorsInPopup: BOOL,
           messagesInPopup: BOOL,
           hideStatusNoMail: BOOL,
           moveToTrash: BOOL,
           newMessagesInWindow: BOOL,
           openInCurrentTab: BOOL,
           dataDirectory: CHAR,
           fullEdit: BOOL,
           font: CHAR },

  export: function() {
    var storage = SmUtils.getThisWindow().browserWindow.SimpleMail.storage;
    var accounts = storage.getAccounts();
    var addresses = storage.getAddresses();

    window.options = { mode: "export",
                       accounts: { checked: accounts.length > 0,
                                   disabled: accounts.length == 0 },
                       passwords: { checked: false,
                                    disabled: accounts.length == 0 },
                       addressBook: { checked: addresses.length > 0,
                                      disabled: addresses.length == 0 },
                       general: { checked: true, disabled: false } };

    SmExport.showOptionsDialog();

    if (options && (options.accounts.checked ||
        options.addressBook.checked || options.general.checked))
    {
      var file = SmExport.chooseFile("save");
      if (!file.leafName.match(/\./)) file.leafName += ".sm"; // If no extension, then append ".sm"
      if (file) SmExport.doExport(file, accounts, addresses);
    }
  },

  showOptionsDialog: function() {
    window.openDialog("chrome://simplemail/content/import/options.xul",
                      "simplemail-preferences-import-options-window",
                      "chrome,centerscreen,modal=yes,resizable=yes");
  },

  chooseFile: function(mode) {
    var fp = Components.classes["@mozilla.org/filepicker;1"]
             .createInstance(Components.interfaces.nsIFilePicker);
    fp.init(window, null, mode == "save" ? fp.modeSave : fp.modeOpen);
    fp.appendFilter("*.sm", "*.sm");
    fp.appendFilters(fp.filterAll);

    var result = fp.show();
    if (result == fp.returnOK || result == fp.returnReplace) return fp.file;
  },

  doExport: function(file, accounts, addresses) {
    var prefs = new Object();

    if (options.accounts.checked) {
      for(var i = 0; i < accounts.length; i++) {
        var account = accounts[i];
        if (options.passwords.checked) {
          account.pop3.getPassword();
          account.smtp.getPassword();
        } else {
          account.pop3.password = null;
          account.smtp.password = null;
        }
      }
      prefs.accounts = accounts;
    }
    if (options.addressBook.checked) {
      prefs.addressBook = addresses;
    }
    if (options.general.checked) {
      prefs.general = new Object();
      for(var name in SmExport.prefs) {
        var type = SmExport.prefs[name];
        var value = (type == CHAR) ? SmPrefs.getCharPref(name) :
                    (type == BOOL) ? SmPrefs.getBoolPref(name)
                                   : SmPrefs.getIntPref(name);
        prefs.general[name] = value;
      }
    }
    SmFile.writeFile(file, SmUtils.toJSON(prefs, "\r\n"), true);
  },

  import: function() {
    var file = SmExport.chooseFile("open");
    if (!file) return;

    var prefs = eval(SmFile.readFile(file, true));

    window.options = { mode: "import",
                       accounts: { checked: prefs.accounts,
                                   disabled: !prefs.accounts },
                       passwords: { checked: prefs.accounts &&
                                             prefs.accounts[0].pop3.password != null,
                                    disabled: !prefs.accounts ||
                                              prefs.accounts[0].pop3.password == null },
                       addressBook: { checked: prefs.addressBook,
                                      disabled: !prefs.addressBook },
                       general: { checked: prefs.general,
                                  disabled: !prefs.general } };

    SmExport.showOptionsDialog();

    if (options && (options.accounts.checked ||
        options.addressBook.checked || options.general.checked))
    {
      SmExport.doImport(prefs);
      return true;
    }
  },

  doImport: function(prefs) {
   var storage = SmUtils.getThisWindow().browserWindow.SimpleMail.storage;

   if (options.accounts.checked) {
      for(var i in prefs.accounts) {
        var account = prefs.accounts[i];
        account.id = null;
        if (!options.passwords.checked) {
          account.pop3.password = null;
          account.smtp.password = null;
        }
        storage.saveAccount(account);
        if (account.color) {
          var canvas = document.createElementNS("http://www.w3.org/1999/xhtml", "canvas");
          canvas.width = 16;
          canvas.height = 16;
          SmIcon.drawUnread(canvas, account.color);
          SmFile.saveCanvas(canvas, SmIcon.getUnreadName(account.color));
          SmIcon.drawRead(canvas, account.color);
          SmFile.saveCanvas(canvas, SmIcon.getReadName(account.color));
        }
      }
    }
    if (options.addressBook.checked) {
      for(var i in prefs.addressBook) {
        var address = prefs.addressBook[i];
        address.id = null;
        storage.saveAddress(address);
      }
    }
    if (options.general.checked) {
      for(var name in prefs.general) {
        var type = SmExport.prefs[name];
        var value = prefs.general[name];
        switch(type) {
          case CHAR: SmPrefs.setCharPref(name, value);
                     break;
          case BOOL: SmPrefs.deletePref(name); // In case old pref of type Int exists
                     SmPrefs.setBoolPref(name, value);
                     break;
          case INT:  SmPrefs.setIntPref(name, value);
                     break;
        }
      }
    }
  }
}