var SmThunderbird = {
  import: function() {
    var fp = Components.classes["@mozilla.org/filepicker;1"]
             .createInstance(Components.interfaces.nsIFilePicker);
    fp.init(window, SmBundle.getString("selectThunderbirdPrefsFile"), fp.modeOpen);
    fp.appendFilter("prefs.js", "prefs.js");
    if (fp.show() == fp.returnOK) {
      SmThunderbird.showEncodingDialog();

      var accounts = SmThunderbird.parsePrefs(SmFile.readFile(fp.file), encoding);

      var file = fp.file.clone();
      file.leafName = "abook.mab";
      var personal = SmThunderbird.parseAddressBook(SmFile.readFile(file), encoding);
      file.leafName = "history.mab";
      var collected = SmThunderbird.parseAddressBook(SmFile.readFile(file), encoding);

      window.options = { mode: "import",
                         accounts: { checked: accounts.length > 0,
                                     disabled: accounts.length == 0 },
                         passwords: { checked: false,
                                      disabled: true },
                         addressBook: { checked: personal.length + collected.length > 0,
                                        disabled: personal.length + collected.length == 0 },
                         general: { checked: false,
                                    disabled: true } };

      SmThunderbird.showOptionsDialog();

      if (options && (options.accounts.checked || options.addressBook.checked)) {
        SmThunderbird.doImport(accounts, personal, collected);
        return true;
      }
    }
  },

  showOptionsDialog: function() {
    window.openDialog("chrome://simplemail/content/import/options.xul",
                      "simplemail-preferences-import-options-window",
                      "chrome,centerscreen,modal=yes,resizable=yes");
  },

  showEncodingDialog: function() {
    window.openDialog("chrome://simplemail/content/import/encoding.xul",
                      "simplemail-preferences-import-encoding-window",
                      "chrome,centerscreen,modal=yes,resizable=yes");
  },

  doImport: function(accounts, personal, collected) {
    if (options.accounts.checked) {
      for(var i = 0; i < accounts.length; i++) {
        storage.saveAccount(accounts[i]);
      }
      SmUtils.error("accountsWereImported", [accounts.length]);
    }
    if (options.addressBook.checked) {
      for(var i in personal) {
        storage.saveAddress(personal[i]);
      }
      for(var i in collected) {
        storage.saveAddress(collected[i]);
      }
    }
  },

  parsePrefs: function(prefs, encoding) {
    function getArray(name) {
      var regexp = new RegExp('"mail\\.' + name + '\\.([^.]+)\\.([^"]+)", ["]?([^")]+)', "g");
      var results = new Array();
      var line;
      while(line = regexp.exec(prefs)) {
        var id = line[1];
        if (!results[id]) results[id] = new Object();
        results[id][line[2]] = line[3];
      }
      return results;
    }

    var identities = getArray("identity");
    var servers = getArray("server");
    var smtpServers = getArray("smtpserver");
    var accounts = getArray("account");
    var results = new Array();

    var defaultSmtp = /"mail\.smtp\.defaultserver", "([^"]+)"/.exec(prefs);
    if (defaultSmtp) defaultSmtp = defaultSmtp[1];

    for(var id in accounts) {
      var server = servers[accounts[id].server];
      if (server.type != "pop3") continue;

      var identityIds = accounts[id].identities.split(",");
      for(var i = 0; i < identityIds.length; i++) {
        var identity = identities[identityIds[i]];
        var smtpServer = smtpServers[identity.smtpServer];
        if (!smtpServer) smtpServer = smtpServers[defaultSmtp];

        var account = SmAccount.create();
        account.name = identity.fullName;
        if (encoding) account.name = SmUtils.toUTF8(account.name, encoding);
        account.email = identity.useremail;

        if (identity.sig_file) {
          var file = identity.sig_file;
          if (encoding) file = SmUtils.toUTF8(file, encoding);
          account.signature = SmFile.readFile(file);
          if (encoding) account.signature = SmUtils.toUTF8(account.signature, encoding)
        }

        account.pop3.host = server.hostname;
        if (server.port) account.pop3.port = server.port;
        account.pop3.ssl = server.socketType == 3;
        account.pop3.login = server.realuserName || server.userName;
        account.smtp.host = smtpServer.hostname;
        if (smtpServer.port) account.smtp.port = smtpServer.port;
        account.smtp.ssl = smtpServer.try_ssl == 3;
        account.smtp.login = smtpServer.username;

        results.push(account);
      }
    }
    return results;
  },

  parseAddressBook: function(addressBook, encoding) {
    function decode(string) {
      var nationalChars = false;
      string = string.replace(/\$(..)|\\\r\n/g, function($0, $1) {
                                                  if ($0 != "\\\r\n") {
                                                    nationalChars = true;
                                                    return String.fromCharCode("0x" + $1);
                                                   }
                                                   else return "";
                                                });
      if (nationalChars && encoding) {
        string = SmUtils.toUTF8(string, encoding);
      }
      return string;
    }
    var addresses = new Array();
    var fields = { name: "DisplayName", email: "PrimaryEmail" };
    var field, ref, data;

    for(var name in fields) {
      var fieldExp = new RegExp("\\(([^=]*)\\s*=(" + fields[name] + ")\\)", "g");

      while(field = fieldExp.exec(addressBook)) {
        var refExp = new RegExp("\\[([^(]*)[^\\]]*\\(\\^" + field[1] +
                                "\\^([^)]*)\\)[^\\]]*\\]", "g");
        while(ref = refExp.exec(addressBook)) {
          var id = ref[1];
          var dataExp = new RegExp("<[^>]*\\(" + ref[2] + "\\s*=([^)]*)\\)[^>]*>", "g");

          while(data = dataExp.exec(addressBook)) {
            if (!addresses[id]) addresses[id] = new Object();
            addresses[id][name] = decode(data[1]);
          }
        }
      }
    }
    return addresses;
  }
}