function SmAddress(id, name, email, sentDate) {
  this.id = id;
  this.name = name;
  this.email = email;
  this.sentDate = sentDate;

  this.getLogin = function() {
    return /(.*)@/.exec(this.email)[1];
  }
  this.getHost = function() {
    return /@(.*)/.exec(this.email)[1];
  }
  this.toString = function() {
    return (this.name ? this.name + " " : "") + "<" + this.email + ">";
  }
  this.toHtml = function() {
    return (this.name ? this.name + " " : "") +
           "&lt;<a href='mailto:" + this.email + "'>" + this.email + "</a>&gt;";
  }
}

SmAddress.parse = function(text) {
  var regexp = /((.*[\s<])?)([^<\s]+@[^>\s]+)/;
  var result = regexp.exec(text);
  return result ? new SmAddress(null, result[1].replace(/^[\s'"<>]*|[\s'"<>]*$/g, ""), result[3]) : null; // Trim spaces & quotes
}

SmAddress.parseList = function(text) {
  var addresses = new Array();
  if (text) {
    var list = text.split(/[,;]/);
    for(var i in list) {
      var address = SmAddress.parse(list[i]);
      if (address) addresses.push(address);
    }
  }
  return addresses;
}

SmAddress.create = function() {
  return new SmAddress(null, null, null, null);
}

var SmIcon = {
  drawEnvelope: function(canvas, color, shadow, fill) {
    var context = canvas.getContext("2d");
    context.clearRect(0, 0, 16, 16);

    // Draw envelope shadow
    context.strokeStyle = shadow;
    context.strokeRect(1.5, 3.5, 14, 10);
    context.beginPath();
    context.moveTo(3, 4);
    context.lineTo(8, 9);
    context.moveTo(7, 9.5);
    context.lineTo(9, 9.5);
    context.moveTo(9, 9);
    context.lineTo(14, 4);
    context.stroke();

    if (fill) {
      context.fillStyle = fill;
      context.fillRect(0.5, 2.5, 14, 10);
    }

    context.strokeStyle = color;
    context.strokeRect(0.5, 2.5, 14, 10);
    context.beginPath();
    context.moveTo(1, 3);
    context.lineTo(7, 9);
    context.moveTo(6, 8.5);
    context.lineTo(9, 8.5);
    context.moveTo(8, 9);
    context.lineTo(14, 3);
    context.stroke();
  },

  drawUnread: function(canvas, color) {
    SmIcon.drawEnvelope(canvas, "black", "silver", color);
  },

  drawRead: function(canvas, color) {
    SmIcon.drawEnvelope(canvas, "gray", "lightgray", color);
  },

  getUnreadName: function(color) {
    return color ? color.substr(1) + "_unread.png" : null;
  },

  getReadName: function(color) {
    return color ? color.substr(1) + "_read.png" : null;
  }
}

function SmAccount(id, name, email, refresh, deleteMode, color, signature, isImap,
                   pop3host, pop3port, pop3ssl, pop3login,
                   smtpHost, smtpPort, smtpSsl, smtpLogin) {
  this.id = id;
  this.name = name;
  this.email = email;
  this.refresh = refresh;
  this.deleteMode = deleteMode;
  this.color = color;
  this.signature = signature;
  this.pop3 = new Object();
  this.pop3.isImap = isImap;
  this.pop3.host = pop3host;
  this.pop3.port = pop3port;
  this.pop3.ssl = pop3ssl;
  this.pop3.login = pop3login;

  // Password is retrieved "on-demand", because if user uses FF Master Password,
  // he is asked by FF for master password when getPassword() method is called.
  // So we want to call getPassword() method only when we really need the password,
  // not when we're just getting a list of accounts from storage.

  this.pop3.getPassword = function() {
    if (this.password == null) {
      this.password = SmPasswordManager.getPassword(id + "/pop3/");
    }
    return this.password;
  }
  this.smtp = new Object();
  this.smtp.host = smtpHost;
  this.smtp.port = smtpPort;
  this.smtp.ssl = smtpSsl;
  this.smtp.login = smtpLogin;
  this.smtp.getPassword = function() {
    if (this.password == null) {
      this.password = SmPasswordManager.getPassword(id + "/smtp/");
    }
    return this.password;
  }

  this.getFullEmail = function() {
    return (this.name ? this.name + " " : "") + "<" + this.email + ">";
  }

  this.getName = function() {
    return this.getFullEmail();
  }

  this.getUnreadIcon = function() {
    return SmIcon.getUnreadName(this.color);
  }

  this.getReadIcon = function() {
    return SmIcon.getReadName(this.color);
  }
}

SmAccount.create = function() {
  return new SmAccount(null,
                       null,
                       null,
                       60,
                       null,
                       null,
                       null,
                       0,
                       null, 
                       110,
                       null,
                       null,
                       null,
                       25,
                       null,
                       null);
}

SmAccount.DELETE_ASK_CONFIRMATION = "ask";
SmAccount.DELETE_ALWAYS = "always";
SmAccount.DELETE_NEVER = "never";

function SmMessage(id, accountId, uid, from, to, cc, bcc, subject, date, size, charset,
                   read, deleted, attachmentsDir, html, folderId, partial, template) {
  this.id = id;
  this.accountId = accountId;
  this.uid = uid;
  this.from = from;
  this.to = to;
  this.cc = cc;
  this.bcc = bcc;
  this.subject = subject;
  this.date = date;
  this.size = size;
  this.charset = charset;
  this.read = read;
  this.deleted = deleted;
  this.attachmentsDir = attachmentsDir || new Date().getTime();
  this.html = html;
  this.folderId = folderId;
  this.partial = partial;
  this.template = template;

  this.contentType;
  this.source;

  this.getSize = function() {
    if (this.size == null) {
      this.size = this.toString().length;
    }
    return this.size;
  }

  this.toString = function() {
    if (!this.source) {
      this.source = SmMessageEncoder.encode(this);
    }
    return this.source;
  }

  this.createReply = function(all) {
    var reply = new SmMessage();
    reply.accountId = this.accountId;
    reply.charset = this.charset;
    reply.to = this.from;
    if (all) {
      if (this.to) reply.to += ", " + this.to;
      reply.cc = this.cc;
      reply.bcc = this.bcc;
    }
    reply.subject = this.subject.replace(/^(Re: )?/i, "Re: ");
    // Remove attachments
    reply.html = this.html.replace(SmMessageEncoder.REGEXP_ATTACHMENT,
                                   function($0, $1, $2) {
                                     return "[Attachment: " + SmFile.getFileName($1 || $2) + "]";
                                   });

    reply.html = "<br><br><span style='color: gray; font-style: italic;'>" +
                 new Date(this.date).toLocaleString() + ", " +
                 SmAddress.parse(this.from).toHtml() + "</span>:<br>" +
                 "<blockquote style='border-left: 1px solid rgb(204, 204, 204);" +
                                             "margin: 0 0 0 0.8ex;" +
                                             "padding: 1ex 0 0 1ex;'>" + reply.html + "</blockquote>\r\n";
    return reply;
  }

  this.createForward = function() {
    function mailList2html(value) {
      var addresses = SmAddress.parseList(value);
      var result = "";
      for(i in addresses) {
        result += (result ? ", " : "") + addresses[i].toHtml();
      }
      return result;
    }
    var forward = new SmMessage();
    forward.accountId = this.accountId;
    forward.charset = this.charset;
    forward.subject = this.subject.replace(/^(Fwd?: )?/i, "Fwd: ");
    forward.html = "<br><br><span style='color: gray; font-style: italic;'>" +
                   "---------- Forwarded message ----------<br>" +
                   "From: " + SmAddress.parse(this.from).toHtml() + "<br>" +
                   "Date: " + new Date(this.date).toLocaleString() + "<br>" +
                   "Subject: " + this.subject + "<br>" +
                   "To: " + mailList2html(this.to) + "<br>" + 
                   (this.cc ? "Cc: " + mailList2html(this.cc) + "<br>" : "") +
                   "</span><br>" + this.html;
    return forward;
  }

  this.equals = function(message) {
    return this.from == message.from &&
           this.to == message.to &&
           this.cc == message.cc &&
           this.bcc == message.bcc &&
           this.charset == message.charset &&
           this.subject == message.subject &&
           this.html == message.html;
  }

  this.copyTo = function(message) {
    message.from = this.from;
    message.to = this.to;
    message.cc = this.cc;
    message.bcc = this.bcc;
    message.charset = this.charset;
    message.subject = this.subject;
    message.html = this.html;
  }
}

SmMessage.parse = function(source) {
  return SmMessageEncoder.decode(source);
}

SmMessage.parseMailto = function(mailto) {
  function getUrlParameter(name) {
    var regexp = new RegExp("[?&]" + name + "=([^&]*)", "i");
    var result = regexp.exec(mailto);
    return result ? decodeURIComponent(result[1]) : null;
  }
  var message = new SmMessage();
  message.to = /mailto:([^?&]*)/.exec(mailto)[1];
  message.cc = getUrlParameter("cc");
  message.bcc = getUrlParameter("bcc");
  message.subject = getUrlParameter("subject");
  message.html = getUrlParameter("body");
  return message;
}

function SmFolder(id, parentId, name) {
  this.id = id;
  this.parentId = parentId;
  this.name = name;
}
SmFolder.DEFAULT_NAMES = ["inbox", "sent", "drafts", "trash"];
SmFolder.INBOX = 1;
SmFolder.SENT = 2;
SmFolder.DRAFTS = 3;
SmFolder.TRASH = 4;
