function SmSocket(host, port, ssl, onClose) {
  var self = this;
  var errors = false;

  this.error = function(messageId, params) {
    errors = true;
    var hostInfo = "[" + host + ":" + port + (ssl ? ",ssl" : "") + "]";
    SmUtils.error(messageId, params ? [hostInfo].concat(params) : [hostInfo]);
  }

  var transportService = Components.classes["@mozilla.org/network/socket-transport-service;1"]
                         .getService(Components.interfaces.nsISocketTransportService);
  var transport = transportService.createTransport(ssl ? ["ssl"] : null, ssl ? 1 : 0, host, port, getProxyInfo());
  var outStream = transport.openOutputStream(0, 1024*1024, 1);

  var inStream = transport.openInputStream(0, 1024*1024, 1);
  var bStream = Components.classes["@mozilla.org/binaryinputstream;1"]
                .createInstance(Components.interfaces.nsIBinaryInputStream);
  bStream.setInputStream(inStream);

  function getProxyInfo() {
    var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService);

    if (prefs.getIntPref("network.proxy.type") == 1) { // Manual proxy configuration
      var proxyHost = prefs.getCharPref("network.proxy.socks");
      if (proxyHost) {
        var proxyPort = prefs.getIntPref("network.proxy.socks_port");
        var proxyService = Components.classes["@mozilla.org/network/protocol-proxy-service;1"]
                           .getService(Components.interfaces.nsIProtocolProxyService);
        return proxyService.newProxyInfo("socks", proxyHost, proxyPort, 0, 30, null);
      }
    }
  }

  var listener = {
    response: "",
    onStartRequest: function(request, context) {},
    onStopRequest: function(request, context, status) {
      bStream.close();
      inStream.close();
      outStream.close();
      if (status) self.error("connectionFailed");
      if (onClose) onClose(status | errors);
    },
    onDataAvailable: function(request, context, inputStream, offset, count) {
      this.response += bStream.readBytes(count);
      if (!this.command || !this.response.match(this.command.responseEnd)) {
        // No command to handle response or response hasn't been read fully, wait
        return;
      }
      var response = this.response;
      this.response = "";

      if (response.match(this.command.responseStart)) {
        if (this.command.onResponse) {
          this.command.onResponse(response);
        }
      }
      else {
        self.error("invalidResponse", [response.replace(/\r\n$/, "")]);
        self.close();
      }
    }
  }

  var pump = Components.classes["@mozilla.org/network/input-stream-pump;1"]
             .createInstance(Components.interfaces.nsIInputStreamPump);
  pump.init(inStream, -1, -1, 0, 0, false);
  pump.asyncRead(listener, null);

  // "responseStart" - regexp for matching correct response status
  // "responseEnd" - regexp for matching response end (to identify whether response is read fully)
  this.Command = function(request, onResponse, responseStart, responseEnd) {
    this.request = request;
    this.onResponse = onResponse;
    this.responseStart = responseStart;
    this.responseEnd = responseEnd;
  }

  this.execute = function(command) {
    listener.command = command;
    if (command.request) {
      var request = command.request;

      function write() {
        try {
          var count = outStream.write(request, request.length);
          if (count < request.length) {
            request = request.substr(count);
            outStream.QueryInterface(Components.interfaces.nsIAsyncOutputStream);
            outStream.asyncWait({ onOutputStreamReady: function() {
                                    write();
                                  }}, 0, 0, null); 
          }
          else outStream.write("\r\n", 2);
        }
        catch(e) { self.error("connectionFailed"); }
      }
      write();      
    }
  }

  this.close = function() {
    transport.close(0);
  }
}

// TODO: SmPOP3, SmSMTP and SmIMAP should not be derived from SmSocket

function SmPOP3(host, port, ssl, onClose) {
  var self = this;
  this.inheritFrom = SmSocket;
  this.inheritFrom(host, port, ssl, onClose);

  this.POP3Command = function(request, onResponse, responseStart, responseEnd) {
    this.inheritFrom = self.Command;
    this.inheritFrom(request, onResponse, responseStart || /^\+OK/,
                                          responseEnd || /\r\n$/);
  }

  this.start = function() {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom();
  }

  this.user = function(login) {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom("USER " + login);
  }

  this.pass = function(password) {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom("PASS " + password);
  }

  this.stat = function() {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom("STAT", function(response) {
      var regexp = /^\+OK\s+(\d+)\s+(\d+)/;
      var count = regexp.exec(response)[1];
      return parseInt(count);
    });
  }

  function GetInfo(request) {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom(request, function(response) {
      var regexp = /\r\n(\S+)\s+(\S+)/g;
      var value, values = new Array();
      while(value = regexp.exec(response)) {
        values[value[1]] = value[2];
      }
      return values;
    },
    null,
    /\r\n\.\r\n$/);
  }

  this.uidl = function() {
    this.inheritFrom = GetInfo;
    this.inheritFrom("UIDL");
  }

  this.list = function() {
    this.inheritFrom = GetInfo;
    this.inheritFrom("LIST");
  }

  function Retrieve(request) {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom(request, function(response) {
      return response.replace(/^.*\r\n|\r\n\.\r\n$/g, "");
    },
    null,
    /\r\n\.\r\n$/);
  }

  this.top = function(id, lines) {
    this.inheritFrom = Retrieve;
    this.inheritFrom("TOP " + id + " " + lines);
  }

  this.retr = function(id) {
    this.inheritFrom = Retrieve;
    this.inheritFrom("RETR " + id);
  }

  this.dele = function(id) {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom("DELE " + id);
  }

  this.quit = function() {
    this.inheritFrom = self.POP3Command;
    this.inheritFrom("QUIT");
  }
}

function SmSMTP(host, port, ssl, onClose) {
  var self = this;
  this.inheritFrom = SmSocket;
  this.inheritFrom(host, port, ssl, onClose);

  this.SMTPCommand = function(request, onResponse, responseStart, responseEnd) {
    this.inheritFrom = self.Command;
    this.inheritFrom(request, onResponse, responseStart || /^2/,
                                          responseEnd || /\r\n$/);
  }

  this.SMTPAuthCommand = function(request, onResponse, responseStart, responseEnd) {
    var command = this;
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom(request, function(response) {
      return response.match(/^504/) ? self.SMTPAuthCommand.NOT_SUPPORTED :
             response.match(command.responseStart) ? self.SMTPAuthCommand.OK :
                                                     self.SMTPAuthCommand.ERROR;
    }, /^/);
  }
  this.SMTPAuthCommand.OK = 1;
  this.SMTPAuthCommand.NOT_SUPPORTED = 2;
  this.SMTPAuthCommand.ERROR = 3;

  this.start = function() {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom();
  }

  function getHostName() {
    var dns = Components.classes["@mozilla.org/network/dns-service;1"]
              .getService(Components.interfaces.nsIDNSService);
    return dns.myHostName;
  }

  this.ehlo = function() {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom("EHLO " + getHostName(), null, null, /2\d* .*\r\n$/);
  }

  this.authPlain = function(login, password) {
    this.inheritFrom = self.SMTPAuthCommand;
    this.inheritFrom("AUTH PLAIN " + btoa("\0" + login + "\0" + password));
  }

  this.authLogin = function() {
    this.inheritFrom = self.SMTPAuthCommand;
    this.inheritFrom("AUTH LOGIN", null, /^334/);
  }

  this.authLoginLogin = function(login) {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom(btoa(login), null, /^334/);
  }

  this.authLoginPassword = function(password) {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom(btoa(password));
  }

  this.mail = function(email) {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom("MAIL FROM:" + email);
  }

  this.rcpt = function(email) {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom("RCPT TO:" + email);
  }

  this.data = function() {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom("DATA", null, /^354/);
  }

  this.dataSend = function(text) {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom(text + "\r\n.");
  }

  this.quit = function() {
    this.inheritFrom = self.SMTPCommand;
    this.inheritFrom("QUIT");
  }
}

function SmIMAP(host, port, ssl, onClose) {
  var self = this;
  this.inheritFrom = SmSocket;
  this.inheritFrom(host, port, ssl, onClose);

  var tag = 1;

  this.IMAPCommand = function(request, onResponse, responseStart, responseEnd) {
    this.inheritFrom = self.Command;
    this.inheritFrom(tag + " " + request, onResponse,
                     new RegExp("(^|\r\n)" + tag + " OK"),
                     new RegExp("(^|\r\n)" + tag + " "));
    tag++;
  }

  this.start = function() {
    this.inheritFrom = self.Command;
    this.inheritFrom(null, null, /\* OK/, /\r\n/);
  }

  this.login = function(login, password) {
    this.inheritFrom = self.IMAPCommand;
    this.inheritFrom("LOGIN " + login + " " + password);
  }

  this.select = function(folder) {
    this.inheritFrom = self.IMAPCommand;
    this.inheritFrom("SELECT " + folder, function(response) {
      var unseen = response.match(/\[UNSEEN (\S+)\]/);
      return unseen && parseInt(unseen[1]);
    });
  }

  this.fetchList = function() {
    this.inheritFrom = self.IMAPCommand;
    this.inheritFrom("FETCH 1:* (UID FLAGS)",
                     function(response) {
                       var regexp = /UID (\S+) FLAGS \((.*?)\)/g;
                       var match, uids = new Array();
                       while(match = regexp.exec(response)) {
                         var flags = match[2];
                         if (!flags.match(/Seen|Deleted/)) uids.push(match[1]);
                       }
                       return uids;
                     });
  }

  this.fetchBody = function(uid) {
    this.inheritFrom = self.IMAPCommand;
    this.inheritFrom("UID FETCH " + uid + " BODY[]",
                     function(response) {
                       return response.replace(/^.*\r\n|\)?\r\n.*\r\n.*\r\n$/g, "");
                     });
  }

  this.storeFlags = function(uids, flags) {
    this.inheritFrom = self.IMAPCommand;
    this.inheritFrom("UID STORE " + uids.join(",") + " +FLAGS ("+ flags + ")",
                     function(response) {
                       var regexp = /UID ([^\s)]+)/g;
                       var match, uids = new Array();
                       while(match = regexp.exec(response)) {
                         uids.push(match[1]);
                       }
                       return uids;
                     });
  }

  this.logout = function() {
    this.inheritFrom = self.IMAPCommand;
    this.inheritFrom("LOGOUT");
  }
}

function SmScenario(socket) {
  var self = this;
  this.scenario = new Array();

  this.executeNextCommand = function() {
    var action = this.scenario.shift();

    if (action) {
      var command = action[0];
      var processResult = action[1];
      var onResponse = command.onResponse;
      command.onResponse = function(response) {
        var result = onResponse ? onResponse(response) : null;
        if (processResult) processResult(result, response);
        self.executeNextCommand();
      };
      socket.execute(command);
    }
    else socket.close();
  }

  this.execute = function() {
    this.executeNextCommand();
    return self;
  }

  this.stop = function() {
    socket.close();
  }
}

function SmGetMail(account, listener) {
  this.inheritFrom = account.pop3.isImap ? Sm_IMAP_GetMail : Sm_POP3_GetMail;
  this.inheritFrom(account, listener);
}

function SmDeleteMail(account, messages, listener) {
  this.inheritFrom = account.pop3.isImap ? Sm_IMAP_DeleteMail : Sm_POP3_DeleteMail;
  this.inheritFrom(account, messages, listener);
}

function Sm_POP3_GetMail(account, listener) {
  // Load password, user will be asked by FF for master password, if any
  account.pop3.getPassword();

  var self = this;
  var pop3 = new SmPOP3(account.pop3.host,
                        account.pop3.port, 
                        account.pop3.ssl,
                        function(status) {
                          listener.onFinish(status);
                        });

  this.inheritFrom = SmScenario;
  this.inheritFrom(pop3);

  this.uids = new Array();

  this.scenario = [ [ new pop3.start(),                     null ],
                    [ new pop3.user(account.pop3.login),    null ],
                    [ new pop3.pass(account.pop3.password), null ],
                    [ new pop3.stat(),                      processStat ],
                    [ new pop3.quit(),                      null ] ];

  function processStat(count) {
    if (count) {
      self.scenario.unshift([ new pop3.uidl(), processUidl ]);
    }
  }

  function processUidl(uids) {
    for(id in SmUtils.reverse(uids)) {
      if (listener.messageExists(account.id, uids[id])) continue;

      self.uids.unshift(uids[id]);
      self.scenario.unshift([ new pop3.retr(id), processRetr ]);
    }
  }

  function processRetr(text) {
    listener.onNewMessage(account.id, self.uids.shift(), text);
  }
}

function Sm_POP3_DeleteMail(account, messages, listener) {
  // Load password, user will be asked by FF for master password, if any
  account.pop3.getPassword();

  var self = this;
  var pop3 = new SmPOP3(account.pop3.host,
                        account.pop3.port,
                        account.pop3.ssl,
                        function(status) {
                          listener.onFinish(status, self.deleted);
                        });

  this.inheritFrom = SmScenario;
  this.inheritFrom(pop3);

  this.ids = new Array();
  this.deleted = new Array();

  this.scenario = [ [ new pop3.start(),                     null ],
                    [ new pop3.user(account.pop3.login),    null ],
                    [ new pop3.pass(account.pop3.password), null ],
                    [ new pop3.uidl(),                      processUidl ],
                    [ new pop3.quit(),                      null ] ];

  function processUidl(uids) {
    var notFound = false;

    function getId(uids, uid) {
      for (var id in uids) {
        if (uids[id] == uid) return id;
      }
    }

    for(var i in messages) {
      var message = messages[i];
      var id = getId(uids, message.uid);
      if (id) {
        self.ids.unshift(message.id);
        self.scenario.unshift([ new pop3.dele(id), processDele ]);
      }
      else notFound = true;
    }
    if (notFound) pop3.error("messagesNotFound");
  }

  function processDele() {
    var id = self.ids.shift();
    self.deleted[id] = messages[id];
    self.deleted[id].uid = null;
  }
}

function SmSendMail(account, message, listener) {
  // Load password, user will be asked by FF for master password, if any
  account.smtp.getPassword();

  var self = this;
  var smtp = new SmSMTP(account.smtp.host,
                        account.smtp.port,
                        account.smtp.ssl,
                        function(status) {
                          listener.onFinish(status);
                        });

  this.inheritFrom = SmScenario;
  this.inheritFrom(smtp);

  var to = SmAddress.parseList(message.to + "," + message.cc + "," + message.bcc);
  message.bcc = ""; // Do not send Bcc header

  this.scenario = [ [ new smtp.start(),                         null ],
                    [ new smtp.ehlo(),                          tryAuthPlain ],
                    [ new smtp.mail("<" + account.email + ">"), prepareRcpt ],
                    [ new smtp.data(),                          null ],
                    [ new smtp.dataSend(message),               null ],
                    [ new smtp.quit(),                          null ] ];

  function tryAuthPlain() {
    if (account.smtp.login) {
      self.scenario.unshift([ new smtp.authPlain(account.smtp.login,
                                                 account.smtp.password), processAuthPlain ]);
    }
  }

  function processAuthPlain(status, response) {
    switch(status) {
      case smtp.SMTPAuthCommand.NOT_SUPPORTED: return tryAuthLogin();
      case smtp.SMTPAuthCommand.ERROR: smtp.error("invalidResponse", [response]);
                                       return smtp.close();
    }
  }

  function tryAuthLogin() {
    self.scenario.unshift([ new smtp.authLogin(), processAuthLogin ]);
  }

  function processAuthLogin(status, response) {
    if (status == smtp.SMTPAuthCommand.OK) {
      self.scenario.unshift([ new smtp.authLoginLogin(account.smtp.login),       null ],
                            [ new smtp.authLoginPassword(account.smtp.password), null ]);
    } else {
      smtp.error("invalidResponse", [response]);
      smtp.close();
    }
  }

  function prepareRcpt() {
    for(var i in SmUtils.reverse(to)) {
      self.scenario.unshift([ new smtp.rcpt("<" + to[i].email + ">"), null ]);
    }
  }
}

function Sm_IMAP_GetMail(account, listener) {
  // Load password, user will be asked by FF for master password, if any
  account.pop3.getPassword();

  var self = this;
  var imap = new SmIMAP(account.pop3.host,
                        account.pop3.port, 
                        account.pop3.ssl,
                        function(status) {
                          listener.onFinish(status);
                        });

  this.inheritFrom = SmScenario;
  this.inheritFrom(imap);

  this.uids = new Array();

  this.scenario = [ [ new imap.start(),                                          null ],
                    [ new imap.login(account.pop3.login, account.pop3.password), null ],
                    [ new imap.select("Inbox"),                         processSelect ],
                    [ new imap.logout(),                                         null ] ];

  function processSelect(unseen) {
    if (unseen) {
      self.scenario.unshift([ new imap.fetchList(), processFetchList ]);
    }
  }

  function processFetchList(uids) {
    for(i in SmUtils.reverse(uids)) {
      self.uids.unshift(uids[i]);
      self.scenario.unshift([ new imap.fetchBody(uids[i]), processFetchBody ]);
    }
  }

  function processFetchBody(text) {
    var uid = self.uids.shift();
    listener.onNewMessage(account.id, uid, text);
    self.scenario.unshift([ new imap.storeFlags([uid], "\\Seen"), null ]);
  }
}

function Sm_IMAP_DeleteMail(account, messages, listener) {
  // Load password, user will be asked by FF for master password, if any
  account.pop3.getPassword();

  var self = this;
  var imap = new SmIMAP(account.pop3.host,
                        account.pop3.port,
                        account.pop3.ssl,
                        function(status) {
                          listener.onFinish(status, self.deleted);
                        });

  this.inheritFrom = SmScenario;
  this.inheritFrom(imap);

  this.deleted = new Array();
  this.uids = new Array();
  for(var i in messages) this.uids.push(messages[i].uid);

  this.scenario = [ [ new imap.start(),                                          null ],
                    [ new imap.login(account.pop3.login, account.pop3.password), null ],
                    [ new imap.select("Inbox"),                                  null ],
                    [ new imap.storeFlags(self.uids, "\\Deleted"),     processDeleted ],
                    [ new imap.logout(),                                         null ] ];

  function processDeleted(uids) {
    var notFound = false;

    function contains(uids, uid) {
      for (var i in uids) {
        if (uids[i] == uid) return true;
      }
    }

    for(var id in messages) {
      if (contains(uids, messages[i].uid)) {
        self.deleted[id] = messages[id];
        self.deleted[id].uid = null;
      }
      else notFound = true;
    }
    if (notFound) imap.error("messagesNotFound");
  }
}
