function SmPasswordManagerImpl() { // Firefox 2.0
  var passwordManager = Components.classes["@mozilla.org/passwordmanager;1"]
                        .getService(Components.interfaces.nsIPasswordManager);
  var passwordManagerInternal = Components.classes["@mozilla.org/passwordmanager;1"]
                                .getService(Components.interfaces.nsIPasswordManagerInternal);

  var url = "chrome://simplemail/";
  var userName = "simplemail";

  this.getPassword = function(id) {
    var host = { value: url + id };
    var user = { value: userName };
    var password = { value: "" }; 
    try {
      passwordManagerInternal.findPasswordEntry(host.value, user.value, password.value, host, user, password);
    }
    catch(e) {}
    return password.value;
  }

  this.removePassword = function(id) {
    try {
      passwordManager.removeUser(url + id, userName);
    } catch(e) {}
  }

  this.setPassword = function(id, password) {
    this.removePassword(id);
    try {
      passwordManager.addUser(url + id, userName, password);
    } catch(e) {}
  }
}

function SmLoginManagerImpl() { // Firefox 3.0
  var loginManager = Components.classes["@mozilla.org/login-manager;1"]
                     .getService(Components.interfaces.nsILoginManager);

  var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1",
                                               Components.interfaces.nsILoginInfo,
                                               "init");
  var url = "chrome://simplemail/";
  var realm = "Simple Mail Account Password";
  var userName = "simplemail";

  this.getPassword = function(id) {
    var logins = loginManager.findLogins({}, url + id, realm, null);
      
    for (var i = 0; i < logins.length; i++) {
      if (logins[i].username == userName) {
         return logins[i].password;
      }
    }
    return "";
  }

  this.removePassword = function(id) {
    var logins = loginManager.findLogins({}, url + id, realm, null);
      
    for (var i = 0; i < logins.length; i++) {
      if (logins[i].username == userName) {
         loginManager.removeLogin(logins[i]);
         return;
      }
    }
  }

  this.setPassword = function(id, password) {
    this.removePassword(id);
    try {    
      var loginInfo = new nsLoginInfo(url + id, realm, null, userName, password, null, null);
      loginManager.addLogin(loginInfo);
    } catch(e) {}
  }
}

var SmPasswordManager = Components.classes["@mozilla.org/passwordmanager;1"]
                        ? new SmPasswordManagerImpl()
                        : new SmLoginManagerImpl();



function SmStorage() {
  var file = SmFile.getDataDirectory();
  file.append("simplemail.sqlite");
  this.firstRun = !file.exists();

  var storageService = Components.classes["@mozilla.org/storage/service;1"]
                       .getService(Components.interfaces.mozIStorageService);
  var connection = storageService.openDatabase(file);

  if (this.firstRun) {
    connection.executeSimpleSQL("CREATE TABLE addressBook (" +
                                  "id             INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "name           STRING," +
                                  "email          STRING," +
                                  "sentDate       DATETIME)");

    connection.executeSimpleSQL("CREATE INDEX addressBook_email ON addressBook (email)");
    connection.executeSimpleSQL("CREATE INDEX addressBook_sentDate ON addressBook (sentDate)");

    connection.executeSimpleSQL("CREATE TABLE accounts (" + 
                                  "id             INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "name           STRING," +
                                  "email          STRING," +
                                  "refresh        DOUBLE," +
                                  "deleteMode     STRING," +
                                  "color          STRING," +
                                  "signature      STRING," +
                                  "isImap         BOOLEAN," +
                                  "pop3host       STRING," +
                                  "pop3port       INTEGER," +
                                  "pop3ssl        BOOLEAN," +
                                  "pop3login      STRING," +
                                  "smtpHost       STRING," +
                                  "smtpPort       INTEGER," +
                                  "smtpSsl        BOOLEAN," +
                                  "smtpLogin      STRING)");

    connection.executeSimpleSQL("CREATE TABLE messages (" + 
                                  "id             INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "accountId      INTEGER," +
                                  "uid            STRING," +
                                  "fromm          STRING," + // "from" is reserved word
                                  "too            STRING," + // "to" is reserved word
                                  "cc             STRING," + 
                                  "bcc            STRING," + 
                                  "subject        STRING," + 
                                  "date           DATETIME," + 
                                  "size           DOUBLE," + 
                                  "charset        STRING," +
                                  "read           BOOLEAN," +
                                  "deleted        BOOLEAN," +
                                  "attachmentsDir STRING," +
                                  "html           STRING," +
                                  "folderId       INTEGER," +
                                  "partial        BOOLEAN," +
                                  "template       BOOLEAN)");

    connection.executeSimpleSQL("CREATE INDEX messages_accountId_uid ON messages (accountId, uid)");
    connection.executeSimpleSQL("CREATE INDEX messages_deleted_folderId_id ON messages (deleted, folderId, id DESC)");

    connection.executeSimpleSQL("CREATE TABLE folders (" + 
                                  "id       INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "parentId INTEGER DEFAULT 0," +
                                  "name     STRING)");
    // Create default folders
    var statement = connection.createStatement("INSERT INTO folders (name) VALUES (?1)");
    for(var i in SmFolder.DEFAULT_NAMES) {
      statement.bindStringParameter(0, SmFolder.DEFAULT_NAMES[i]);
      statement.execute();
    }
  }

  try { // Since version 2.00
    connection.executeSimpleSQL("ALTER TABLE messages ADD bcc STRING");
    connection.executeSimpleSQL("ALTER TABLE messages ADD folderId INTEGER DEFAULT " + SmFolder.INBOX); // Set Inbox as a folder for all legacy messages
    connection.executeSimpleSQL("ALTER TABLE messages ADD partial BOOLEAN");
    connection.executeSimpleSQL("ALTER TABLE messages ADD template BOOLEAN");
    connection.executeSimpleSQL("DROP INDEX messages_deleted_id");
    connection.executeSimpleSQL("CREATE INDEX messages_deleted_folderId_id ON messages (deleted, folderId, id DESC)");
  }
  catch(e) {}

  try { // Since version 2.02
    connection.executeSimpleSQL("ALTER TABLE addressBook ADD sentDate DATETIME");
  }
  catch(e) {}

  try { // Since version 2.03
    connection.executeSimpleSQL("ALTER TABLE accounts ADD color STRING");
  }
  catch(e) {}

  try { // Since version 2.13
    connection.executeSimpleSQL("CREATE TABLE folders (" + 
                                  "id       INTEGER PRIMARY KEY AUTOINCREMENT," +
                                  "parentId INTEGER DEFAULT 0," +
                                  "name     STRING)");
    // Create default folders
    var statement = connection.createStatement("INSERT INTO folders (name) VALUES (?1)");
    for(var i in SmFolder.DEFAULT_NAMES) {
      statement.bindStringParameter(0, SmFolder.DEFAULT_NAMES[i]);
      statement.execute();
    }
  }
  catch(e) {}

  try { // Since version 2.31
    connection.executeSimpleSQL("ALTER TABLE accounts ADD isImap BOOLEAN");
  }
  catch(e) {}

  var text = "text: ";

  function bindString(statement, id, value) {
    // Force 'text' type for strings with leading zeroes, otherwise SQLite strips these zeroes
    statement.bindStringParameter(id, (value && value[0] == "0") ? text + value : value);
  }

  function getString(statement, id) {
    var value = statement.getString(id);
    return (value && value.substr(0, text.length) == text) ? value.substr(text.length) : value;
  }

  function getAddress(email) {
    var sql = "SELECT id, name, email, sentDate FROM addressBook WHERE email=?1";
    var statement = connection.createStatement(sql);
    try {
      bindString(statement, 0, email);
      return statement.executeStep() ? new SmAddress(statement.getInt32(0),
                                                     getString(statement, 1),
                                                     getString(statement, 2),
                                                     statement.getInt64(3))
                                     : null;
    }
    finally {
      statement.reset();
    }
  }

  this.saveAddress = function(address) {
    // If address already exists, update it
    var exists = getAddress(address.email);
    if (exists) address.id = exists.id;

    var sql = address.id ? "UPDATE addressBook SET name=?1, email=?2, sentDate=?3 WHERE id=?4"
                         : "INSERT INTO addressBook (name, email, sentDate) VALUES (?1, ?2, ?3)";

    var statement = connection.createStatement(sql);

    bindString(statement, 0, address.name);
    bindString(statement, 1, address.email);
    statement.bindInt64Parameter(2, new Date().getTime());
    if (address.id) statement.bindInt32Parameter(3, address.id);
    statement.execute();

    if (!address.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM addressBook");
      try {
        if (statement.executeStep()) address.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
  }

  this.addAddressList = function(list) {
    var addresses = SmAddress.parseList(list);
    for(var i in addresses) {
      this.saveAddress(addresses[i]);
    }
  }

  function getAddresses(sql) {
    var statement = connection.createStatement(sql);
    try {
      var addresses = new Array();
      while(statement.executeStep()) {
        addresses.push(new SmAddress(statement.getInt32(0),
                                     getString(statement, 1),
                                     getString(statement, 2),
                                     statement.getInt64(3)));
      }
      return addresses;
    }
    finally {
      statement.reset();
    }
  }

  this.getAddresses = function() {
    return getAddresses("SELECT id, name, email, sentDate FROM addressBook ORDER BY name, email");
  }                               

  this.getRecentAddresses = function() {
    return getAddresses("SELECT id, name, email, sentDate FROM addressBook ORDER BY sentDate DESC LIMIT 3");
  }

  this.deleteAddress = function(address) {
    var statement = connection.createStatement("DELETE FROM addressBook WHERE id=?1");
    statement.bindInt32Parameter(0, address.id);
    statement.execute();
  }

  var accounts; // Accounts list cache

  this.getAccounts = function() {
    if (!accounts) {
      accounts = doGetAccounts();
    }
    return accounts;
  }

  function doGetAccounts() {
    var accounts = new Array();
    var sql = "SELECT id, name, email, refresh, deleteMode, color, signature, isImap," +
              "pop3host, pop3port, pop3ssl, pop3login, " +
              "smtpHost, smtpPort, smtpSsl, smtpLogin FROM accounts";
    var statement = connection.createStatement(sql);
    try {
      while(statement.executeStep()) {
        var id = statement.getInt32(0);
        accounts.push(new SmAccount(id,
                                    getString(statement, 1),
                                    getString(statement, 2),
                                    statement.getDouble(3),
                                    getString(statement, 4),
                                    getString(statement, 5),
                                    getString(statement, 6),
                                    statement.getInt32(7),
                                    getString(statement, 8),
                                    statement.getInt32(9),
                                    statement.getInt32(10),
                                    getString(statement, 11),
                                    getString(statement, 12),
                                    statement.getInt32(13),
                                    statement.getInt32(14),
                                    getString(statement, 15)));
      }
    }
    finally {
      statement.reset();
    }
    return accounts;
  }

  this.saveAccount = function(account) {
    var sql = account.id ? "UPDATE accounts SET name=?1, email=?2, refresh=?3, deleteMode=?4, color=?5, signature=?6, isImap=?7," + 
                           "pop3host=?8, pop3port=?9, pop3ssl=?10, pop3login=?11, " +
                           "smtpHost=?12, smtpPort=?13, smtpSsl=?14, smtpLogin=?15 WHERE id=?16"

                         : "INSERT INTO accounts (name, email, refresh, deleteMode, color, signature, isImap," + 
                           "pop3host, pop3port, pop3ssl, pop3login, " + 
                           "smtpHost, smtpPort, smtpSsl, smtpLogin) " +
                           "VALUES (?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, ?15)";

    var statement = connection.createStatement(sql);

    bindString(statement, 0, account.name);
    bindString(statement, 1, account.email);
    statement.bindDoubleParameter(2, account.refresh);
    bindString(statement, 3, account.deleteMode);
    bindString(statement, 4, account.color);
    bindString(statement, 5, account.signature);
    statement.bindInt32Parameter(6, account.pop3.isImap);
    bindString(statement, 7, account.pop3.host);
    statement.bindInt32Parameter(8, account.pop3.port);
    statement.bindInt32Parameter(9, account.pop3.ssl);
    bindString(statement, 10, account.pop3.login);
    bindString(statement, 11, account.smtp.host);
    statement.bindInt32Parameter(12, account.smtp.port);
    statement.bindInt32Parameter(13, account.smtp.ssl);
    bindString(statement, 14, account.smtp.login);
    if (account.id) statement.bindInt32Parameter(15, account.id);
    statement.execute();

    if (!account.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM accounts");
      try {
        if (statement.executeStep()) account.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
    SmPasswordManager.setPassword(account.id + "/pop3/", account.pop3.password);
    SmPasswordManager.setPassword(account.id + "/smtp/", account.smtp.password);

    // Invalidate accounts list cache
    accounts = null;
  }

  this.deleteAccount = function(account) {
    var statement = connection.createStatement("DELETE FROM accounts WHERE id=?1");
    statement.bindInt32Parameter(0, account.id);
    statement.execute();

    SmPasswordManager.removePassword(account.id + "/pop3/");
    SmPasswordManager.removePassword(account.id + "/smtp/");

    if (account.color) {
      SmFile.deleteImage(account.getUnreadIcon());
      SmFile.deleteImage(account.getReadIcon());
    }

    // Invalidate accounts list cache
    accounts = null;
  }

  this.messageExists = function(accountId, uid) {
    var statement = connection.createStatement("SELECT uid FROM messages WHERE accountId=?1 AND uid=?2");
    try {
      statement.bindInt32Parameter(0, accountId);
      bindString(statement, 1, uid);
      return statement.executeStep();
    }
    finally {
      statement.reset();
    }
  }

  this.getMessages = function(folderId, count, order) {
    var sql = "SELECT id, accountId, uid, fromm, too, cc, bcc, subject, date, size, " +
              "charset, read, deleted, attachmentsDir, folderId, partial, template " +
              "FROM messages WHERE deleted = 0 AND folderId = ?1 ORDER BY " +
              (order || "id DESC") +
              (count ? " LIMIT ?2" : "");
    var statement = connection.createStatement(sql);
    try {
      var messages = new Array();
      statement.bindInt32Parameter(0, folderId);
      if (count) statement.bindInt32Parameter(1, count);

      while(statement.executeStep()) {
        var id = statement.getInt32(0);
        messages.push(new SmMessage(id,
                                    statement.getInt32(1),
                                    getString(statement, 2),
                                    getString(statement, 3),
                                    getString(statement, 4),
                                    getString(statement, 5),
                                    getString(statement, 6),
                                    getString(statement, 7),
                                    new Date(statement.getInt64(8)),
                                    statement.getDouble(9),
                                    getString(statement, 10),
                                    statement.getInt32(11),
                                    statement.getInt32(12),
                                    getString(statement, 13),
                                    null,
                                    statement.getInt32(14),
                                    statement.getInt32(15),
                                    statement.getInt32(16)));
      }
      return messages;
    }
    finally {
      statement.reset();
    }
  }

  this.saveMessage = function(message) {
    // Do not really delete message, keep its uid in DB so that it's not downloaded from server again
    if (message.deleted) {
      SmFile.deleteAttachments(message.attachmentsDir);

      message.from = null;
      message.to = null;
      message.cc = null;
      message.subject = null;
      message.date = null;
      message.charset = null;
      message.read = false;
      message.attachmentsDir = null;
      message.html = null;
      message.partial = false;
      message.template = false;
    }
    else SmFile.saveAttachment(message.attachmentsDir,
                               "message.txt",
                               message.source || message.toString());

    var sql = message.id ? "UPDATE messages SET accountId=?1, uid=?2, fromm=?3, too=?4, cc=?5, " +
                           "bcc=?6, subject=?7, date=?8, size=?9, charset=?10, read=?11, deleted=?12, " +
                           "attachmentsDir=?13, html=?14, folderId=?15, partial=?16, template=?17 WHERE id=?18"
                         : "INSERT INTO messages (accountId, uid, fromm, too, cc, bcc, subject, date, " +
                           "size, charset, read, deleted, attachmentsDir, html, folderId, partial, template) " +
                           "VALUES (?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, ?15, ?16, ?17)";

    var statement = connection.createStatement(sql);

    statement.bindInt32Parameter(0, message.accountId);
    bindString(statement, 1, message.uid);
    bindString(statement, 2, message.from);
    bindString(statement, 3, message.to);
    bindString(statement, 4, message.cc);
    bindString(statement, 5, message.bcc);
    bindString(statement, 6, message.subject);
    statement.bindInt64Parameter(7, Date.parse(message.date));
    statement.bindDoubleParameter(8, message.getSize());
    bindString(statement, 9, message.charset);
    statement.bindInt32Parameter(10, message.read);
    statement.bindInt32Parameter(11, message.deleted);
    bindString(statement, 12, message.attachmentsDir);
    bindString(statement, 13, message.html);
    statement.bindInt32Parameter(14, message.folderId);
    statement.bindInt32Parameter(15, message.partial);
    statement.bindInt32Parameter(16, message.template);
    if (message.id) statement.bindInt32Parameter(17, message.id);
    statement.execute();

    if (!message.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM messages");
      try {
        if (statement.executeStep()) message.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
  }

  this.getMessage = function(id) {
    var sql = "SELECT id, accountId, uid, fromm, too, cc, bcc, subject, date, size, " +
              "charset, read, deleted, attachmentsDir, html, folderId, partial, template " +
              "FROM messages WHERE id=?1";      
    var statement = connection.createStatement(sql);
    try {
      statement.bindInt32Parameter(0, id);
      if (statement.executeStep()) return new SmMessage(id,
                                                        statement.getInt32(1),
                                                        getString(statement, 2),
                                                        getString(statement, 3),
                                                        getString(statement, 4),
                                                        getString(statement, 5),
                                                        getString(statement, 6),
                                                        getString(statement, 7),
                                                        new Date(statement.getInt64(8)),
                                                        statement.getDouble(9),
                                                        getString(statement, 10),
                                                        statement.getInt32(11),
                                                        statement.getInt32(12),
                                                        getString(statement, 13),
                                                        getString(statement, 14),
                                                        statement.getInt32(15),
                                                        statement.getInt32(16),
                                                        statement.getInt32(17));
    }
    finally {
      statement.reset();
    }
  }

  this.saveMessageState = function(message) {
    var statement = connection.createStatement("UPDATE messages SET read=?1, folderId=?2 WHERE id=?3");
    statement.bindInt32Parameter(0, message.read);
    statement.bindInt32Parameter(1, message.folderId);
    statement.bindInt32Parameter(2, message.id);
    statement.execute();
  }

  this.deleteMessage = function(message) {
    if (message.uid) {
      message.deleted = true;
      this.saveMessage(message);
    } else {
      SmFile.deleteAttachments(message.attachmentsDir);
      var statement = connection.createStatement("DELETE FROM messages WHERE id=?1");
      statement.bindInt32Parameter(0, message.id);
      statement.execute();
    }
  }

  this.getFolders = function() {
    var statement = connection.createStatement("SELECT id, parentId, name FROM folders");
    try {
      var folders = new Array();
      while(statement.executeStep()) {
        var parentId = statement.getInt32(1);
        if (!folders[parentId]) {
          folders[parentId] = new Array();
        }
        folders[parentId].push(new SmFolder(statement.getInt32(0),
                                            parentId,
                                            getString(statement, 2)));
      }
      return folders;
    }
    finally {
      statement.reset();
    }
  }

  this.saveFolder = function(folder) {
    var sql = folder.id ? "UPDATE folders SET parentId = ?1, name = ?2 WHERE id=?3"
                        : "INSERT INTO folders (parentId, name) VALUES (?1, ?2)";
    var statement = connection.createStatement(sql);
    statement.bindInt32Parameter(0, folder.parentId);
    bindString(statement, 1, folder.name);
    if (folder.id) statement.bindInt32Parameter(2, folder.id);
    statement.execute();

    if (!folder.id) {
      var statement = connection.createStatement("SELECT last_insert_rowid() FROM folders");
      try {
        if (statement.executeStep()) folder.id = statement.getInt32(0);
      }
      finally {
        statement.reset();
      }
    }
  }

  // Deletes single folder with messages (doesn't delete subfoders)
  this.deleteFolder = function(folder) {
    var messages = this.getMessages(folder.id);
    for(var i = 0; i < messages.length; i++) {
      this.deleteMessage(messages[i]);
    }
    var statement = connection.createStatement("DELETE FROM folders WHERE id=?1");
    statement.bindInt32Parameter(0, folder.id);
    statement.execute();
  }
}
