function $(id) {
  return document.getElementById(id);
}

// "Wed Oct 10 2007 14:52:53 GMT+0300" => "Wed, 10 Oct 2007 14:52:53 +0300"
Date.prototype.toMIMEString = function() {
  return this.toString().replace(/(^\w+) (\w+) (\w+) (.*) GMT(.*)/, "$1, $3 $2 $4 $5");
}

var SmUtils = {
  // Make a reversed copy of an array, preserving keys
  reverse: function(array) {
    var reverseKeys = new Array();
    for(var i in array) {
      reverseKeys.unshift(i);
    }
    var result = new Array();
    for(var i in reverseKeys) {
      var key = reverseKeys[i];
      result[key] = array[key];
    }
    return result;
  },

  merge: function(array1, array2) {
    var result = new Array();
    for(var i in array1) {
      result[i] = array1[i];
    }
    for(var i in array2) {
      result[i] = array2[i];
    }
    return result;
  },

  clearElement: function(element) {
    for(var i = element.childNodes.length - 1; i >= 0; i--) {
      element.removeChild(element.childNodes[i]);
    }
  },

  alert: function(message, title) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.alert(window, title || "Simple Mail", message);
  },

  confirm: function(message) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    return prompts.confirm(window, "Simple Mail", message);
  },

  prompt: function(message, value) {
    var prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Components.interfaces.nsIPromptService);
    var message = (SmBundle.getString(message) || message) + ":";
    return prompts.prompt(window, "Simple Mail", message, value, null, { value: true });
  },

  popup: function(image, title, message, onClick) {
    var listener = { observe: function(subject, topic, data) {
                                if (topic == "alertclickcallback" && onClick) {
                                  onClick();
                                }
                              }};
    var alertsService = Components.classes["@mozilla.org/alerts-service;1"]
                        .getService(Components.interfaces.nsIAlertsService);
    alertsService.showAlertNotification(image, title, message, true, null, listener);
  },

  error: function(messageId, params) {
    var simpleMailError = SmBundle.getString("simpleMailError");
    var message = SmBundle.getString(messageId, params);

    if (SmPrefs.getBoolPref("errorsInPopup")) {
      var browserWindow = SmUtils.getThisWindow().browserWindow;
      SmUtils.popup("chrome://simplemail/skin/status/errors_large.png",
                    simpleMailError, message,
                    function() {
                      browserWindow.toJavaScriptConsole();
                    });
    } else {
      SmUtils.alert(message, simpleMailError);
    }

    message = simpleMailError + ": " + message;
    Components.utils.reportError(message);
    return message;
  },

  getThisWindow: function() {
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
             .getService(Components.interfaces.nsIWindowMediator);
    var browserWindow = wm.getMostRecentWindow("navigator:browser");
      return { "window"       : window,
               "browserWindow": browserWindow,
               "browser"      : browserWindow ? browserWindow.getBrowser() : null,
               "tab"          : browserWindow ? browserWindow.getBrowser().selectedTab : null };
  },

  toUTF8: function(text, encoding) {
    try {
      var utf = Components.classes["@mozilla.org/intl/utf8converterservice;1"]
                .getService(Components.interfaces.nsIUTF8ConverterService);
      return utf.convertStringToUTF8(text, encoding, false);
    }
    catch(e) {
      return text;
    }
  },

  fromUTF8: function(text, encoding) {
    try {
      var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"]
                      .getService(Components.interfaces.nsIScriptableUnicodeConverter);
      converter.charset = encoding;
      var result = converter.ConvertFromUnicode(text);
      return result + converter.Finish();
    }
    catch(e) {
      return text;
    }
  },

  // TODO: do conversion through this: [@mozilla.org/streamconv;1?from=text/plain&to=text/html]
  text2html: function(text) {
    return text.replace(/</g, "&lt;")
               .replace(/>/g, "&gt;")
               .replace(/\r\n/g, "<br>")
               .replace(/(https?:|mailto:)[^\s<>\[\]\(\){}]+/g, function($0) {
                 return "<a href='" + $0 + "'>" + $0 + "</a>";
               });
  },

  toJSON: function(object, indent, level) {
    if (!level) level = 0;
    if (!indent) indent = " ";
    var json = "";

    for(var i in object) {
      var value = object[i];
      if (i[0] != "_" &&     // Field names beginning with "_" are ignored
          value != null &&
          typeof value != "function")
      {
        if (json) json += "," + indent;
        json += "'" + i + "': ";

        switch(typeof value) {
          case "number":
          case "boolean":
              json += value;
              break;
          case "string":
              json += "'" + value.replace(/([\\\'\n])/g, "\\$1") + "'";
              break;
          default:
              json += this.toJSON(value, indent, level + 1);
              break;
        }
      }
    }
    json = "{" + indent + json + indent + "}";
    return level ? json : "(" + json + ")";
  },

  getSelectedHtml: function() {
    function link(href, title) {
      if (!href) href = content.document.location.href;
      if (!title) title = content.document.title;
      return "<a href='" + href + "'>" + (title || href) + "</a>";
    }
    var html;
    var selection = content.getSelection();

    if (selection.rangeCount) {
      var fragment = selection.getRangeAt(0).cloneContents();
      var div = content.document.createElement("div");
      div.appendChild(fragment);
      html = div.innerHTML;
    }
    if (!html && gContextMenu.target.nodeName == "IMG") {
      html = "<img src='" + gContextMenu.target.src + "'>";
    }
    if (html) {
      return html + "<br><br>" + SmBundle.getString("source") + ": " + link();
    }
    else if (gContextMenu.target.nodeName == "A") {
      return link(gContextMenu.target.href, gContextMenu.target.innerHTML);
    }
    else return link();
  }
}

var DefaultTreeView = {
  setTree: function(treebox) { this.treebox = treebox; },
  isContainer: function(row) { return false; },
  isSeparator: function(row) { return false; },
  isSorted: function() { return false; },
  cycleHeader: function(column) {},
  cycleCell: function(row, column) { this.treebox.view.selection.select(row); },
  getLevel: function(row) { return 0; },
  getImageSrc: function(row, column) { return null; },
  getRowProperties: function(row, props) {},
  getCellProperties: function(row, column, props) {},
  getColumnProperties: function(column, props) {},
  canDrop: function(row, orientation) { return false; },

  extend: function(treeView) {
    for(var i in this) {
      if (treeView[i] == null) treeView[i] = this[i];
    }
  },

  atomService: Components.classes["@mozilla.org/atom-service;1"]
               .getService(Components.interfaces.nsIAtomService),

  nextDirection: {
    ""           : "ascending",
    "ascending"  : "descending",
    "descending" : ""
  },

  getSortedColumn: function() {
    var column = messageTree.columns.getSortedColumn();
    if (column) {
      return $(column.id);
    }
  },

  forEachSelectedItem: function(onItem) {
    var start = new Object();
    var end = new Object();
    var rangeCount = this.treebox.view.selection.getRangeCount();

    this.treebox.beginUpdateBatch();
    for (var i = rangeCount - 1; i >= 0; i--){
      this.treebox.view.selection.getRangeAt(i, start, end);
      for (var j = end.value; j >= start.value; j--) {
        onItem(j);
      }
    }
    this.treebox.endUpdateBatch();
  }
}

var SmBundle = {
  bundle: Components.classes["@mozilla.org/intl/stringbundle;1"]
          .getService(Components.interfaces.nsIStringBundleService)
          .createBundle("chrome://simplemail/locale/simplemail.properties"),

  getString: function(name, params) {
    try {
      return params ? SmBundle.bundle.formatStringFromName(name, params, params.length) 
                    : SmBundle.bundle.GetStringFromName(name);
    }
    catch(e) {
      return null;
    }
  }
}
